/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 *
 * Revision History
 *	11/25/91	3.1.1	patchlevel 14
 *		Added "login.defs" to Makefile
 *	12/02/91	3.1.2	patchlevel 15
 *		Bugs found by users
 *	12/28/91	3.1.3	patchlevel 16
 *		Changes for SunOS 4.1.1
 *	02/08/92	3.1.4	patchlevel 17
 *		Changes for SVR4, plus bug fixes
 *	04/03/92	3.2.1	patchlevel 18
 *		Minor bug fixes, new baseline
 *	07/07/92	3.2.2	patchlevel 20
 *		Added administrator defined authentication
 *	11/04/92	3.2.3	patchlevel 21
 *		Bug fixes for SVR4
 *	07/23/93	3.3.0	patchlevel 23
 *		New baseline release
 *	08/23/93	3.3.1	patchlevel 24
 *		Bug fixes for SunOS 4.1.1
 */

#define	RELEASE		3
#define	PATCHLEVEL	24
#define	VERSION		"3.3.1"
